/***************************************************************************/
/*                                                                         */
/*  File: Game.cpp                                                         */
/*  Date: 06-02-07                                                         */
/*                                                                         */
/***************************************************************************/
/*
	Main Game File - This is where we update all our elements in the game
	do any creation of things, such as land, tanks and call render
	on them all or update....heart of the whole game is here.

*/
/***************************************************************************/


#include <math.h>		// cosf(..) sinf(..), sqrt(..)
#include <stdio.h>      // sprintf(..)

#include <windows.h>    // opengl has issues if we don't include it!

#include <d3d9.h>
#include <d3dx9.h>					// DirectX Header Files
#pragma comment(lib, "d3d9.lib")	// DirectX Library Files
#pragma comment(lib, "d3dx9.lib")					

#include <stdio.h>					// sprintf(..)

#include "dxdbg.h"					// Various dx debug drawing functions
#include "CubeCollision.h"			// CubeCubeCollisionCheck(...)
#include "Cube.h"


#include "Game.h"



// Externals

// DirectX Device/Render Handles
extern LPDIRECT3D9			g_pD3D;			// main.cpp
extern LPDIRECT3DDEVICE9	g_pD3DDevice;	// main.cpp

// Globals
bool		g_wireView		= true;
float		g_timeStep		= 1.0f;			// Timing 
bool		g_paused		= false;
bool		g_singleStep	= false;
float		g_totalTime		= 0.0f;

int			g_numCubes		= 0;
Cube		g_cubes[10]; // Max cubes for now


// Simple structure to store our collision information
struct stCollisionPoints
{
	Cube* box0;
	Cube* box1;
	D3DXVECTOR3 point;
	D3DXVECTOR3 normal;
	float		pen;
};

// Simple and dirty, but just a big array of collision data
stCollisionPoints g_Collisions[100];
int g_numCols = 0;




/***************************************************************************/

void CreateCubeSetup( int config )
{
	switch (config)
	{
		case 0:
		case 1: // Single Cube On Plane
		{
			Cube cubes[2] = 
				{
					Cube ( D3DXVECTOR3(0,5,0),		D3DXVECTOR3(0,0,0),			D3DXVECTOR3(5,5,5),		1.0f),
					Cube ( D3DXVECTOR3(0,-5.0f,0),	D3DXVECTOR3(0,0,0),			D3DXVECTOR3(50, 10, 50), 100000.0f )
				};
			g_numCubes = sizeof(cubes)/sizeof(Cube);
			memcpy(g_cubes, cubes, sizeof(Cube)*g_numCubes);
		}
		break;

		case 2: // Two Cubes Stacked
		{
			Cube cubes[3] = 
				{
					Cube( D3DXVECTOR3(0,12,0),		D3DXVECTOR3(0,0.4f,0.4f),	D3DXVECTOR3(5,5,5),		1.0f),
					Cube( D3DXVECTOR3(0,5,0),		D3DXVECTOR3(0,0,0),			D3DXVECTOR3(5,5,5),		1.0f),
					Cube ( D3DXVECTOR3(0,-5.0f,0),	D3DXVECTOR3(0,0,0),			D3DXVECTOR3(50, 10, 50),	100000.0f )
				};
			g_numCubes = sizeof(cubes)/sizeof(Cube);
			memcpy(g_cubes, cubes, sizeof(Cube)*g_numCubes);
		}
		break;

		case 3: // Cube Bouncing Off Plank
		{
			Cube cubes[4] = 
				{
					Cube( D3DXVECTOR3(5,12,0),		D3DXVECTOR3(0,0,0.4f),		D3DXVECTOR3(3,3,3),		1.0f),
					Cube( D3DXVECTOR3(0,6,0),		D3DXVECTOR3(0,0,0),			D3DXVECTOR3(20,2,5),	1.0f),
					Cube( D3DXVECTOR3(0,2.5,0),		D3DXVECTOR3(0,0,0),			D3DXVECTOR3(5,5,5),		1.0f),
					Cube ( D3DXVECTOR3(0,-5.0f,0),	D3DXVECTOR3(0,0,0),			D3DXVECTOR3(50, 10, 50), 100000.0f )
				};
			g_numCubes = sizeof(cubes)/sizeof(Cube);
			memcpy(g_cubes, cubes, sizeof(Cube)*g_numCubes);
		}
		break;

		case 4:
		{
			Cube cubes[2] = 
				{
					Cube( D3DXVECTOR3(0,6,0),		D3DXVECTOR3(0,0.4f,0.4f),	D3DXVECTOR3(5,5,5),		1.0f),
					Cube ( D3DXVECTOR3(0,-5.0f,0),	D3DXVECTOR3(0,0,0),			D3DXVECTOR3(50, 10, 50), 100000.0f )
				};
			g_numCubes = sizeof(cubes)/sizeof(Cube);
			memcpy(g_cubes, cubes, sizeof(Cube)*g_numCubes);
		}
		break;

		case 5: // Multiple Cubes Stacked
		{
			Cube cubes[4] = 
				{
					Cube ( D3DXVECTOR3(0,12,0),		D3DXVECTOR3(0,0.1f,0.2f),	D3DXVECTOR3(5,5,5),		1.0f),
					Cube ( D3DXVECTOR3(0,5,0),		D3DXVECTOR3(0,0,0),			D3DXVECTOR3(5,5,5),		1.0f),
					Cube ( D3DXVECTOR3(0,18,0),		D3DXVECTOR3(0,0,0),			D3DXVECTOR3(5,5,5),		1.0f),
					Cube ( D3DXVECTOR3(0,-5.0f,0),	D3DXVECTOR3(0,0,0),			D3DXVECTOR3(50, 10, 50), 100000.0f )
				};
			g_numCubes = sizeof(cubes)/sizeof(Cube);
			memcpy(g_cubes, cubes, sizeof(Cube)*g_numCubes);
		}
		break;

		case 6: // Multiple Cubes Stacked
		{
			Cube cubes[3] = 
				{
					//Cube ( D3DXVECTOR3(0,12.5f,0),	D3DXVECTOR3(0,0,0),			D3DXVECTOR3(5,5,5),		1.0f),
					Cube ( D3DXVECTOR3(0,7.5,0),	D3DXVECTOR3(0,0.1f,0),		D3DXVECTOR3(5,5,5),		1.0f),
					Cube ( D3DXVECTOR3(0,2.5,0),	D3DXVECTOR3(0,0,0),			D3DXVECTOR3(5,5,5),		1.0f),
					Cube ( D3DXVECTOR3(0,-5.0f,0),	D3DXVECTOR3(0,0,0),			D3DXVECTOR3(50, 10, 50), 100000.0f )
				};
			g_numCubes = sizeof(cubes)/sizeof(Cube);
			memcpy(g_cubes, cubes, sizeof(Cube)*g_numCubes);
		}
		break;

		case 7:
		{
			Cube cubes[2] = 
				{
					Cube ( D3DXVECTOR3(0,10,0),		D3DXVECTOR3(0.4f,0.4f,0.4f),	D3DXVECTOR3(5,5,5),			1.0f),
					Cube ( D3DXVECTOR3(0,-5.0f,0),	D3DXVECTOR3(0,0,0),				D3DXVECTOR3(50, 10, 50),	100000.0f )
				};
			g_numCubes = sizeof(cubes)/sizeof(Cube);
			memcpy(g_cubes, cubes, sizeof(Cube)*g_numCubes);
		}
		break;

		case 8: //  Something a little more exciting...a wall of cubes..3x3 high and long
		{
			const int numX = 3;
			const int numY = 3;

			Cube cubes[numX * numY + 1];

			for (int x=0; x<numX; x++)
			{
				for (int y=0; y<numY; y++)
				{
					float w = 5.0f;
					float h = 3.0f;
					float d = 5.0f;
					cubes[x + y*numX] = Cube ( D3DXVECTOR3(x*w,y*h + h/2.0f,0),		D3DXVECTOR3(0,0,0),		D3DXVECTOR3(w,h,d),		1.0f);
				}
			}
			cubes[numX * numY] = Cube ( D3DXVECTOR3(0,-0.5f,0),	D3DXVECTOR3(0,0,0),				D3DXVECTOR3(50, 1, 50), 100000.0f );

			g_numCubes = sizeof(cubes)/sizeof(Cube);
			memcpy(g_cubes, cubes, sizeof(Cube)*g_numCubes);
		}
		break;

		case 9: // Cube on a plane to test sliding friction
		{
			Cube cubes[2] = 
				{
					Cube( D3DXVECTOR3(0,5,0),		D3DXVECTOR3(0,0.4f,0.0f),		D3DXVECTOR3(5,5,5),		1.0f),
					Cube( D3DXVECTOR3(0,0,0),		D3DXVECTOR3(0,0,0.4f),			D3DXVECTOR3(20,1,10),	100000.0f),
				};
			g_numCubes = sizeof(cubes)/sizeof(Cube);
			memcpy(g_cubes, cubes, sizeof(Cube)*g_numCubes);
		}
		break;

	}

	for (int i=0; i<g_numCubes; i++)
	{
		g_cubes[i].UpdateMatrix();
	}
}

/***************************************************************************/

bool Create()
{	
	CreateCubeSetup(2);

	return true;
}// End Create

/***************************************************************************/

bool Release()
{

	return true;
}// End Release()


/***************************************************************************/
/*                                                                         */
/*  Name: KeyBoard() & KeyDown()										   */
/*                                                                         */
/*  Used to accept keyboard input and alter some global variables - used   */
/*  for simple analysis of our code - turning wire frame rendering on/off  */
/*  adjusting variables which are used in the demo.                        */
/*                                                                         */
/***************************************************************************/
static void KeyBoard()
{
	static float step = 0.1f;

	if( GetKeyState(VK_RIGHT)   & 0x80 ) g_cubes[0].m_c.x += step;
	if( GetKeyState(VK_LEFT)    & 0x80 ) g_cubes[0].m_c.x -= step;
	if( GetKeyState(VK_UP)      & 0x80 ) g_cubes[0].m_c.y += step;
	if( GetKeyState(VK_DOWN)    & 0x80 ) g_cubes[0].m_c.y -= step;
	if( GetKeyState('Z')		& 0x80 ) g_cubes[0].m_c.z += step;
	if( GetKeyState('X')		& 0x80 ) g_cubes[0].m_c.z -= step;

	if( GetKeyState('D')		& 0x80 ) g_cubes[0].AddRotateX(0.05f);
	if( GetKeyState('F')		& 0x80 ) g_cubes[0].AddRotateX(-0.05f);
	if( GetKeyState('E')		& 0x80 ) g_cubes[0].AddRotateZ(0.05f);
	if( GetKeyState('R')		& 0x80 ) g_cubes[0].AddRotateZ(-0.05f);

	for (int i=0; i<10; i++)
	{
		if( GetKeyState('0'+i)		& 0x80 ) CreateCubeSetup(i);
		
	}
}//End KeyBoard()

// &&&&&&&&&

// So we can catch single key presses for single time step
void KeyDown()
{
	if ( GetKeyState('P') & 0x80 )	g_paused		= !g_paused;
	if ( GetKeyState('S') & 0x80 )	g_singleStep	= true;
	if ( GetKeyState('W') & 0x80 )	g_wireView		= !g_wireView;
}//KeyDown(..)


/***************************************************************************/


static void UpdateTiming()
{
	// We use a more accurate timing information, QueryPerformanceCounter
	// instead of GetTickCount() so we get smoother animations
	LARGE_INTEGER freq;
	QueryPerformanceFrequency(&freq);

	LARGE_INTEGER nTime;
	QueryPerformanceCounter(&nTime);

	static LARGE_INTEGER oTime = nTime;

	double delay = (double(nTime.QuadPart - oTime.QuadPart)/freq.QuadPart);
	oTime=nTime;

	g_timeStep = (float)delay;

	if (g_timeStep > 0.07f)
	{
		g_timeStep = 0.07f;
	}

}

/***************************************************************************/

void SetCamera()
{
	extern bool  g_LMousePress;
	extern bool  g_RMousePress;
	extern int	 g_mousePosX;
	extern int   g_mousePosY;
	extern int   g_zoom;

	float s = 0.02f;
	D3DXVECTOR3 vMove = D3DXVECTOR3((float)g_mousePosX*s, (float)g_mousePosY*s, 0);

	static D3DXVECTOR3 vDir			= D3DXVECTOR3(0, 20, -50);
	static D3DXVECTOR3 vDirOld		= vDir;


	if (g_LMousePress || g_RMousePress)
	{
		if (g_LMousePress)
		{
			D3DXMATRIX rot;
			D3DXMatrixRotationYawPitchRoll(&rot, vMove.x, vMove.y, 0);
			D3DXVec3TransformCoord(&vDir, &vDirOld, &rot);
		}
		else if (g_RMousePress)
		{
			vDir = vDirOld;
			D3DXVECTOR3 vDirNorm;
			D3DXVec3Normalize(&vDirNorm, &vDir);
			float scale = 1.0f;
			vDir += (float)g_zoom*vDirNorm*scale;
		}
	}
	else
	{
		vDirOld = vDir;
	}

	
	static D3DXVECTOR3 vLookAt		= D3DXVECTOR3(0,0,0);
	D3DXVECTOR3 vCamPosition		= vLookAt + vDir;


    // [1] D3DTS_VIEW
    D3DXMATRIX v;
    g_pD3DDevice->SetTransform(	D3DTS_VIEW, 
								D3DXMatrixLookAtLH(&v, 
								&vCamPosition, 
                                &vLookAt, 
                                &D3DXVECTOR3(0,1,0)));  
    // [2] D3DTS_PROJECTION
    D3DXMATRIX p;
    g_pD3DDevice->SetTransform( D3DTS_PROJECTION, 
								D3DXMatrixPerspectiveFovLH( &p,  
								D3DX_PI/4,  
								1.0f,
                                1.0f,1000.0f));
}//End SetCamera(..)


/***************************************************************************/
/*                                                                         */
/* Update()                                                                */
/* Our main render loop, which gets called over and over agian to do our   */
/* drawing...clears the screen, draws the data, then presents it..         */
/*                                                                         */
/***************************************************************************/
void Update()
{

	UpdateTiming();

	KeyBoard();

	SetCamera();

	
	// Clear the back buffer to a blue color
	g_pD3DDevice->Clear( 0, 
						 NULL, 
						 D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER, 
						 D3DCOLOR_XRGB(0,20,155), 
						 1.0f, 
						 0 );

	g_pD3DDevice->BeginScene();


	static debug_wire_ground ground(g_pD3DDevice, 12,12);
	ground.Draw(g_pD3DDevice, 0,0,0, 50.0f);


	if (g_wireView)	g_pD3DDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_WIREFRAME );
	else			g_pD3DDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_SOLID );



	D3DXVECTOR3 hitGroundPoint;
	D3DXVECTOR3 hitPoint;
	float penetrationBox;
	D3DXVECTOR3 hitNormalBox;
	D3DXVECTOR3 hitNormalGround;
	bool hitBox    = false;
	bool hitGround = false;

	float dt = g_timeStep * 1.0f;

	if (g_paused)
	{
		dt = 0.0f;
	}

	bool doSingleStep = false;
	if (g_singleStep)
	{
		doSingleStep = true;
		g_singleStep = false;
		dt = 0.01f;
	}

	// Things go wierd if our timestep dt is to large!...so clamp it
	dt = Clamp(dt, 0.0f, 0.1f);

	g_totalTime += dt;




	// Check Cube-Cube Collisions
	for (int i=0; i<g_numCubes; i++)
	{
		for (int j=i; j<g_numCubes; j++)
		{
			if (j!=i)
			{
				D3DXVECTOR3 hitPoints[10];
				int numHitPoints = 0;
				hitBox = CubeCubeCollisionCheck(g_cubes[i], 
												g_cubes[j], 
												hitPoints, 
												&numHitPoints, 
												&penetrationBox, 
												&hitNormalBox);

				if (hitBox)
				{
					
					for (int k=0; k<numHitPoints; k++)
					{

						// Draw the collision point!
						static debugsphere sphere(g_pD3DDevice, 0xffffff00);
						sphere.Draw(g_pD3DDevice, hitPoints[k].x, hitPoints[k].y, hitPoints[k].z, 0.4f);

						// Work out if where colliding with an object that doesn't move, such as the
						// earch...I assume if mass is greater than 1000 its imovable
						float invMass0 = g_cubes[i].m_mass > 1000.0f ? 0 : (1.0f/g_cubes[i].m_mass);
						float invMass1 = g_cubes[j].m_mass > 1000.0f ? 0 : (1.0f/g_cubes[j].m_mass);
						float totalInvMass = invMass0 + invMass1;


						// Clamp the penetration depth to a max of 1.0f, don't really need this, but
						// it keeps things stable just incase
						if (penetrationBox > 1.0f)
						{	
							penetrationBox-=1.0f;
							g_cubes[i].m_c += hitNormalBox*(penetrationBox)*(invMass0/totalInvMass);
							g_cubes[j].m_c -= hitNormalBox*(penetrationBox)*(invMass1/totalInvMass);
							penetrationBox=1.0f;
						}

						// Work out the radio of penetration...if its hitting the ground...the ground
						// has infinite mass...so the penetration will be from the ground upwards into the
						// shape...for two normal shapes..its just 50:50 
						D3DXVECTOR3 p0 = hitPoints[k] + hitNormalBox*penetrationBox*(invMass0/totalInvMass);
						D3DXVECTOR3 p1 = hitPoints[k] - hitNormalBox*penetrationBox*(invMass1/totalInvMass);
						
						// Draw the spring!...maybe do a DrawSpringLine3D(..) at some point, as it would look cooler :)
						DrawLine3D(g_pD3DDevice, p0, p1);

						
						// Use good old Hooks law... F = -kx

						D3DXVECTOR3 force0 = (p0 - hitPoints[k]) * 50.0f;
						g_cubes[i].AddForce(hitPoints[k], force0);

						D3DXVECTOR3 force1 = (p1 - hitPoints[k]) * 50.0f;
						g_cubes[j].AddForce(hitPoints[k], force1);
					}
				}
			}
		}
	}


	for (int i=0; i<g_numCubes; i++)
	{
		g_cubes[i].Render(g_pD3DDevice);
	}
	


	for (int i=0; i<g_numCubes; i++)
	{
		g_cubes[i].Update(dt);
	}





	DrawText(g_pD3DDevice, "OBB vs OBB Collsion Demo", 10, 10); 
	char buf[256] = {'\0'};
	sprintf(buf, "Collision: %s", hitBox?"YES":"NO");
	DrawText(g_pD3DDevice, buf, 10, 30);
	sprintf(buf, "Time: %.3f", g_totalTime);
	DrawText(g_pD3DDevice, buf, 10, 50);

	DrawText(g_pD3DDevice, "Keys 1-9 Select Config", 10, 70);
	DrawText(g_pD3DDevice, "P-Pause",				 10, 90);
	DrawText(g_pD3DDevice, "S-SingleStep",			 10, 110);


	g_pD3DDevice->EndScene();
	// After rendering the scene we display it.
	g_pD3DDevice->Present( NULL, NULL, NULL, NULL );

}//End Update()






